#!/usr/bin/perl -w

# friends.pl
my $friends_version = "2.0 (20011228)";

# Maintains list of people you know.
# Rewritten by Jakub Jankowski <shasta@quasimodo.olsztyn.tpsa.pl>
# Based on friends.pl by Erkki Seppl <flux@inside.org> 2001
# get friends.pl-related things at http://irssi.atn.pl/friends/

# This script have been written for Irssi 0.7.98.CVS 20011201+
# It may not work with older ones.

# GNU GPL2 code.

use Irssi;
use Irssi::Irc;
use strict;

# These variables are adjustable with /set
# but here are some 'safe' defaults:

# do you want to process CTCP queries?
my $friends_use_ctcp = 1;

# space-separated list of allowed (implemented ;) CTCP commands
my $friends_ctcp_commands = "OP VOICE LIMIT KEY INVITE PASS";

# do you want to learn new users?
my $friends_learn = 0;

# which flags do you want to use? (case *sensitive*)
my $friends_allowed_flags = "aoOvViklLF";

# which flags do you want to add automatically with /addfriend? (case *sensitive*)
my $friends_default_flags = "";

# path to friendlist
my $friends_file = "$ENV{HOME}/.irssi/friends";

# minimal delay
my $delay_min = 5;

# maximal delay
my $delay_max = 25;

# default debug level
# higher value == more messages
# 0 means off
my $friends_debug_level = 0;

# registering themes
Irssi::theme_register([
	'friends_version', 'friends.pl\'s version:    {hilight $0} [$1]',
	'friends_file_written', 'friendlist written on:   {hilight $0}',
	'friends_file_version', 'friendlist written with: {hilight $0} [$1]',
	'friends_filetooold', 'Friendfile too old, loading aborted',
	'friends_loaded', 'Loaded {hilight $0} friends from $1',
	'friends_saved', 'Saved {hilight $0} friends to $1',
	'friends_duplicate', 'Skipping %R$0%n [duplicate?]',
	'friends_checking', 'Checking {hilight $0} took {hilight $1} secs [on $2]',
	'friends_notenoughargs', 'Not enough arguments. Usage: $0',
	'friends_badargs', 'Bad arguments. Usage: $0',
	'friends_chflagexec', 'Executing %c$0%n for %R$1%n',
	'friends_currentflags', 'Current {channel $2} flags for %R$1%n are: %c$0%n',
	'friends_chpassexec', 'Altered password for %R$0%n',
	'friends_line_head', '[$[!-3]0] Handle: %R$1%n, flags: %C$2%n [password: $3]',
	'friends_line_hosts', '$[-6]9 Hosts: $0',
	'friends_line_chan', '$[-6]9 Channel {hilight $0}: Flags: %c$1%n, Delay: $2',
	'friends_line_currentnick', '$[-6]9 Current nick: {nick $0}',
	'friends_empty', 'Your friendlist is empty. Add items with /ADDFRIEND',
	'friends_friendlist', '{hilight Friendlist} [$0]:',
	'friends_findfriends', 'Online friends on channel {hilight $0} [on $1]:',
	'friends_added', 'Added %R$0%n to friendlist',
	'friends_removed', 'Removed %R$0%n from friendlist',
	'friends_host_added', 'Added {hilight $1} to %R$0%n',
	'friends_host_removed', 'Removed {hilight $1} from %R$0%n',
	'friends_host_exists', 'Hostmask {hilight $1} overlaps with one of the already added to %R$0%n',
	'friends_host_notexists', '%R$0%n does not have {hilight $1} in hostlist',
#	'friends_chanrec_added', 'Added {hilight $1} record to %R$0%n',
#	'friends_chanrec_exists', '%R$0%n has got {hilight $1} record already',
	'friends_chanrec_removed', 'Removed {hilight $1} record from %R$0%n',
	'friends_chanrec_notexists', '%R$0%n does not have {hilight $1} record',
	'friends_changed_handle', 'Changed {hilight $0} to %R$1%n',
	'friends_changed_delay', 'Changed %R$0%n\'s delay value on {hilight $1} to %c$2%n',
	'friends_nosuch', 'No such friend %R$0%n',
	'friends_ctcprequest', '%R$0%n asks for {hilight $1} on {hilight $2}',
	'friends_ctcppass', 'Password for %R$0%n altered by $1',
	'friends_ctcpfail', 'Failed CTCP {hilight $0} from %R$1%n. $2',
	'friends_endof', 'End of $0 $1',
	'friends_notafriend', 'No such friend %R$0%n',
	'friends_notuniqhandle', 'Handle %R$0%n already exists, choose another one',
	'friends_badhandle', 'Wrong handle: %R$0%n. $1',
	'friends_optree_header', 'Opping tree:',
	'friends_optree_line1', '%R$0%n has opped these:',
	'friends_optree_line2', '{hilight $0} times: $1'
]);

# Initialize vars
my @friends = ();
my @operationQueue = ();
my $timerHandle = undef;
my $friends_file_version;
my $friends_file_written;

# void printversion($what)
# print's version of script/userlist
sub printversion {
	my ($what) = @_;
	$what = lc($what);

	if ($what eq "filever") {
		my ($file_verl, $file_vern) = $friends_file_version =~ /^(.+)\ \(([0-9]+)\)$/;
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_file_version', $file_verl, $file_vern);
	} elsif ($what eq "filewritten") {
		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($friends_file_written);
		my $written_string = sprintf("%4d%02d%02d %02d:%02d:%02d", ($year+1900), ($mon+1), $mday, $hour, $min, $sec);
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_file_written', $written_string);
	} else {
		my ($verl, $vern) = $friends_version =~ /^(.+)\ \(([0-9]+)\)$/;
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_version', $verl, $vern);
	}
}

# bool globMatcher ($what, $regex)
# glob matching
sub globMatcher ($$) {
	my ($what, $regex) = @_;

	debug("globMatcher(): what='$what', regex='$regex'", 7);

	$regex =~ s/\\/\\\\/g;
	$regex =~ s/\./\\\./g;
	$regex =~ s/\*/\.\*/g;
	$regex =~ s/\!/\\\!/g;
	# Lam's suggestion: '?' means "any, but always one, character"
	$regex =~ s/\?/\./g;
	$regex =~ s/\+/\\\+/g;
	$regex =~ s/\^/\\\^/g;
	$regex =~ s/\$/\\\$/g;
	$regex =~ s/\[/\\\[/g;
	$regex =~ s/\]/\\\]/g;
	$regex =~ s/\(/\\\(/g;
	$regex =~ s/\)/\\\)/g;
	$regex =~ s/\|/\\\|/g;
	my $matches = $what =~ /^$regex$/i;

	debug("globMatcher(): after quoting: what='$what', regex='$regex'", 7);
	debug("globMatcher(): -- ".(($matches) ? "%_MATCHED%_" : "Not matched" ), 7);

	return $matches;
}

# void debug($text, $level)
# prints debug message if friends_debug_level >= $level
sub debug {
	my ($text, $level) = @_;
	my $debug_level = Irssi::settings_get_int('friends_debug_level');
	Irssi::print("DEBUG: ".$text) if ($debug_level >= $level);
}

# str friends_crypt($plain)
# returns crypt()ed $plain, using random salt;
# or "" if $plain is empty
sub friends_crypt {
	my ($plain) = @_;
	if ($plain eq "") {
		debug("friends_crypt(): got empty password", 2);
		return;
	}
	my $crypted = crypt("$plain", (join '', ('.', '/', 0..9, 'A'..'Z', 'a'..'z')[rand 64, rand 64]));
	return $crypted;
}

# bool friend_passwdok($idx, $pwd)
# returns 1 if password is ok, 0 if isn't
sub friends_passwdok {
	my ($idx, $pwd) = @_;
	if (crypt("$pwd", $friends[$idx]->{password}) eq $friends[$idx]->{password}) {
		debug("friends_passwdok(): Password for $friends[$idx]->{handle} is OK", 3);
		return 1;
	}
	debug("friends_passwdok(): Password for $friends[$idx]->{handle} not matched", 3);
	return 0;
}

# arr get_friends_channels($idx)
# returns list of $friends[$idx] channels
sub get_friends_channels {
	my ($idx) = @_;
	return keys(%{$friends[$idx]->{channels}});
}

# arr get_friends_hosts($idx);
# returns list of $friends[$idx] hosts
sub get_friends_hosts {
	my ($idx) = @_;
	return split(" ", $friends[$idx]->{hosts});
}

# str get_friends_flags($idx, $chan)
# returns list of $chan flags for $idx
# $chan might be '' or 'global' too
# case insensitive about the $chan
sub get_friends_flags {
	my ($idx, $chan) = @_;
	$chan = lc($chan);
	if ($chan eq "" || $chan eq "global") {
		return $friends[$idx]->{globflags};
	} else {
		foreach my $friendschan (get_friends_channels($idx)) {
			if ($chan eq lc($friendschan)) {
				return $friends[$idx]->{channels}->{$friendschan}->{flags};
			}
		}
	}
	debug("get_friends_flags() failed for idx=$idx, chan=$chan", 5);
	return;
}

# str get_friends_delay($idx, $chan)
# returns $chan delay for $idx
# returns "" if $chan equals '' or 'global'
# case insensitive about the $chan
sub get_friends_delay {
	my ($idx, $chan) = @_;
	$chan = lc($chan);
	if ($chan eq "" || $chan eq "global") {
		return;
	} else {
		foreach my $friendschan (get_friends_channels($idx)) {
			if ($chan eq lc($friendschan)) {
				return $friends[$idx]->{channels}->{$friendschan}->{delay};
			}
		}
	}
	debug("get_friends_delay() failed for idx=$idx, chan=$chan", 5);
	return;
}

# void renumber_friendlist()
# updates every $friends[$idx]->{number}
sub renumber_friendlist {
	for (my $idx = 0; $idx < @friends; ++$idx) {
		$friends[$idx]->{number} = $idx;
	}
	debug("renumber_friendlist(): Done renumbering", 4);
}


# struct friend new_friend($handle, $hosts, $globflags, $chanstr, $password)
# chanstr is: #chan1,flags,delay #chan2,flags,delay
sub new_friend {
	# $friend's structure:
	# $friend->{number} = number
	# $friend->{handle} = handle
	# $friend->{hosts} = *!foo@host1 *!bar@host2 *!?baz@host3
	# $friend->{globflags} = oikl [global-flags]
	# $friend->{channels}->{#chan1}->{exist} = 1
	# $friend->{channels}->{#chan1}->{flags} = dk
	# $friend->{channels}->{#chan1}->{delay} = 5
	# $friend->{channels}->{#chan2}->{exist} = 1
	# $friend->{channels}->{#chan2}->{flags} = ao
	# $friend->{channels}->{#chan2}->{delay} = 
	# $friend->{password} = crypt(password)
	# $friend->{friends} = []

	my $friend = {};
	$friend->{number} = scalar(@friends);
	$friend->{handle} = shift;
	$friend->{hosts} = shift;
	$friend->{globflags} = shift;
	my $chanstr = shift;
	$friend->{password} = shift;
	$friend->{friends} = [];

	my @chanflags = split(" ", $chanstr);
	foreach my $chanflag (@chanflags) {
		# $entry[0] == #foobar (channelname)
		# $entry[1] == oikl (flags)
		# $entry[2] == 15 (delay)
		my @entry = split(",", $chanflag);
		my $channel = $entry[0];
		$friend->{channels}->{$channel}->{exist} = 1;
		$friend->{channels}->{$channel}->{flags} = $entry[1];
		$friend->{channels}->{$channel}->{delay} = $entry[2];
	}

	return $friend;
}

# bool is_allowed_flag($flag)
# checks if $flag is in range of friends_allowed_flags
sub is_allowed_flag {
	my ($flag) = @_;
	my $allowed = Irssi::settings_get_str('friends_allowed_flags');
	if ($allowed =~ /$flag/) {
		return 1;
	}
	debug("is_allowed_flag(): Flag $flag is out of allowed flags range ($allowed)", 5);
	return 0;
}

# bool is_ctcp_command($command)
# check if $command is one of the implemented ctcp commands
sub is_ctcp_command {
	my ($command) = @_;
	my $ctcp_commands = Irssi::settings_get_str('friends_ctcp_commands');
	foreach my $allowed (split(/[,\ ]+/, $ctcp_commands)) {
		if (uc($command) eq uc($allowed)) {
			return 1;
		}
	}
	debug("is_ctcp_command(): Command $command isn't one of the implemented (ctcp_commands)", 4);
	return 0;
}

# int get_idx($channelName, $nick, $userhost)
# returns idx of the friend or -1 if not a friend
sub get_idx {
	my ($chan, $nick, $userhost) = @_;
	$chan = lc($chan);
	# browse through all friends
	for (my $idx = 0; $idx < @friends; ++$idx) {
		if ((friend_has_chanrec($idx, $chan) || get_friends_flags($idx, "global") ne "") &&
			friend_has_host($idx, $nick."!".$userhost)) {
			debug("get_idx(): matched $nick!$userhost/$chan to idx=$idx", 6);
			return $idx;
		}
	}
	debug("get_idx(): failed on $nick!$userhost/$chan", 6);
	return -1;
}

# int get_idxbyhand($handle)
# returns $idx of friend with $handle or -1 if no such handle
# case insensitive
sub get_idxbyhand {
	my ($handle) = @_;
	for (my $idx = 0; $idx < @friends; ++$idx) {
		if (lc($friends[$idx]->{handle}) eq lc($handle)) {
			return $idx;
		}
	}
	debug("get_idxbyhand() failed on $handle", 6);
	return -1;
}

# int get_idxbyhost($nick, $userhost)
# returns $idx of friend with first matching $nick!$userhost
# or -1 if no such friend
# can be dangerous if used improperly :)
sub get_idxbyhost {
	my ($nick, $userhost) = @_;
	for (my $idx = 0; $idx < @friends; ++$idx) {
		foreach my $hostmask (get_friends_hosts($idx)) {
			if (globMatcher(($nick."!".$userhost), $hostmask)) {
				return $idx;
			}
		}
	}
	debug("get_idxbyhost() failed on $nick!$userhost", 4);
	return -1;
}

# int friend_has_host($idx, $host)
# checks wheter any of $friend[$idx]'s hosts can be matched with $host
# case sensitive
sub friend_has_host {
	my ($idx, $host) = @_;
	foreach my $hostmask (get_friends_hosts($idx)) {
		# tries to match in both ways; does such way have any flaws?
		if (globMatcher($host, $hostmask) || globMatcher($hostmask, $host)) {
			debug("friend_has_host(): Matched: $host and $hostmask", 6);
			return 1;
		}
	}
	return 0;
}

# void add_host($idx, $host)
# adds $host to $friends[$idx]->{hosts}
sub add_host {
	my ($idx, $host) = @_;
	$friends[$idx]->{hosts} = join(" ", (split(" ", $friends[$idx]->{hosts}), $host));
	debug("add_host(): Added $host to $friends[$idx]->{handle}", 5);
}

# void del_host($idx, $host)
# deletes $host from $friends[$idx]->{hosts}
sub del_host {
	my ($idx, $host) = @_;
	my $tmpstr;
	# if you know a better way for doing this, let me know
	# i'm really not a perl monger ;)
	foreach my $hostmask (get_friends_hosts($idx)) {
		if ($hostmask ne $host) {
			$tmpstr .= $hostmask." ";
		}
	}
	$tmpstr =~ s/[\ ]+$//;
	$friends[$idx]->{hosts} = $tmpstr;
	debug("del_host(): Deleted $host from $friends[$idx]->{handle}", 5);
}

# bool friend_has_chanrec($idx, $chan)
# checks wheter $friend[$idx] has a $chan record
# case insensitive
sub friend_has_chanrec {
	my ($idx, $chan) = @_;
	$chan = lc($chan);
	foreach my $friendschan (get_friends_channels($idx)) {
		if ($chan eq lc($friendschan)) {
			debug("friend_has_chanrec(): $friends[$idx]->{handle} has channel record for $chan", 6);
			return 1;
		}
	}
	debug("friend_has_chanrec(): $friends[$idx]->{handle} doesn't have channel record for $chan", 6);
	return 0;
}

# void add_chanrec($idx, $chan)
# adds an empty $chan record to $friends[$idx]
# case sensitive
sub add_chanrec {
	my ($idx, $chan) = @_;
	$friends[$idx]->{channels}->{$chan}->{exist} = 1;
	debug("add_chanrec(): Added $chan record to $friends[$idx]->{handle}", 5);
#	Irssi::printformat(MSGLEVEL_CRAP, 'friends_chanrec_added', $friends[$idx]->{handle}, $chan);
}

# void del_chanrec($idx, $chan)
# deletes $chan record from $friends[$idx]
# case *in*sensitive
sub del_chanrec {
	my ($idx, $chan) = @_;
	foreach my $friendschan (get_friends_channels($idx)) {
		if (lc($chan) eq lc($friendschan)) {
			delete $friends[$idx]->{channels}->{$friendschan};
			debug("del_chanrec(): Deleted $friendschan record from $friends[$idx]->{handle}", 5);
		}
	}
}

# struct friend delfriend($idx)
# removes friend no. $idx from $friends
# returns the removed $friend object
sub delfriend {
	my ($idx) = @_;
	my $result = -1;
	if ($idx < scalar(@friends)) {
		$result = splice(@friends, $idx, 1);
	}
	debug("delfriend(): Deleted $result->{handle}", 4);
	return $result
}

# bool uniq_handle($handle)
# checks if the $handle is unique for the whole friendlist
# returns 1 if there's no such $handle
# returns 0 if there is one.
sub uniq_handle {
	my ($handle) = @_;
	my $idx = get_idxbyhand($handle);
	if ($idx > -1) {
		debug("uniq_handle() failed on $handle", 2);
		return 0;
	}
	return 1;
}

# str choose_handle($proposed)
# tries to choose a handle, closest to the $proposed one
sub choose_handle {
	my ($proposed) = @_;
	my $counter = 0;
	my $handle = $proposed;

	# do this until we have an unique handle
	while (!uniq_handle($handle)) {
		if (!($handle =~ /([0-9]+)$/) && !$counter) {
			# first, if handle doesn't end with a digit, append '0'
			# (but only in first step)
			$handle .= "0";
		} elsif ($counter < 20) {
			# later, increase the trailing number by one
			# do that 19 times
			my ($number) = $handle =~ /([0-9]+)$/;
			++$number;
			$handle =~ s/([0-9]+)$/$number/;
		} elsif ($counter == 20) {
			# then, if it didn't helped, make $handle = $proposed."_"
			$handle = $proposed . "_";
		} elsif ($counter < 40) {
			# if still unsuccessful, append "_" to the handle
			# do that 19 times
			$handle .= "_";
		} else {
			# if THAT didn't help -- make some silly handle
			# and exit the loop
			$handle = $proposed.'_'.(join '', (0..9, 'a'..'z')[rand 36, rand 36, rand 36, rand 36]);
			last;
		}
		++$counter;
	}

	debug("choose_handle() returning $handle in step: $counter", 2);
	# return our glorious handle ;-)
	return $handle;
}

# bool friend_has_flag($idx, $flag, $chan)
# returns true if $friends[$idx] has $flag for $chan
# (checks global flags, if $chan is 'global' or unspecified)
# returns false if hasn't or $flag is not in allowed_flags
# case sensitive about the FLAG
# case insensitive about the chan.
sub friend_has_flag {
	my ($idx, $flag, $chan) = @_;

	# return false if $flag is out of friends_allowed_flags range
	return 0 if (!is_allowed_flag($flag));

	if (get_friends_flags($idx, $chan) =~ /$flag/) {
		debug("friend_has_flag(): $friends[$idx]->{handle} has $chan flag '$flag'", 6);
		return 1;
	}
	debug("friend_has_flag(): $friends[$idx]->{handle} doesn't have $chan flag '$flag'", 6);
	return 0;
}

# bool addflag($idx, $flag, $chan)
# adds $flag to $idx's $chan flags
# $chan can be '' or 'global'
# case insensitive about the $chan -- chooses the proper case.
# returns 1 on success
sub addflag {
	my ($idx, $flag, $chan) = @_;
	$chan = lc($chan);
	if ($chan eq "" || $chan eq "global") {
		$friends[$idx]->{globflags} .= $flag;
		return 1;
	} else {
		foreach my $friendschan (get_friends_channels($idx)) {
			if ($chan eq lc($friendschan)) {
				$friends[$idx]->{channels}->{$friendschan}->{flags} .= $flag;
				return 1;
			}
		}
	}
	debug("addflag() failed for idx=$idx, flag=$flag, chan=$chan", 3);
	return 0;
}

# bool delflag($idx, $flag, $chan)
# removes $flag from $idx's $chan flags
# $chan can be '' or 'global'
# case insensitive about the $chan -- chooses the proper case.
sub delflag {
	my ($idx, $flag, $chan) = @_;
	$chan = lc($chan);
	if ($chan eq "" || $chan eq "global") {
		$friends[$idx]->{globflags} =~ s/$flag//g;
		return 1;
	} else {
		foreach my $friendschan (get_friends_channels($idx)) {
			if ($chan eq lc($friendschan)) {
				$friends[$idx]->{channels}->{$friendschan}->{flags} =~ s/$flag//i;
				return 1;
			}
		}
	}
	debug("delflag() failed for idx=$idx, flag=$flag, chan=$chan", 3);
	return 0;
}

# bool chdelay($idx, $delay, $chan)
# alters $idx's delay time for $chan
# $chan may not be 'global' or empty
sub chdelay {
	my ($idx, $delay, $chan) = @_;
	$chan = lc($chan);
	if ($chan eq "" || $chan eq "global") {
		debug("chdelay() failed for idx=$idx, delay=$delay, chan=$chan", 3);
		return 0;
	} else {
		foreach my $friendschan (get_friends_channels($idx)) {
			if ($chan eq lc($friendschan)) {
				$friends[$idx]->{channels}->{$friendschan}->{delay} = $delay;
				return 1;
			}
		}
	}
	debug("chdelay() failed for idx=$idx, delay=$delay, chan=$chan", 3);
	return 0;
}

# void listfriend($idx[, $nick])
# prints an info line about certain friend.
# if you want to improve the look of the script, you should
# change /format friends_*, probably.
sub listfriend {
	my ($idx, $nick) = @_;
	my $passset = ($friends[$idx]->{password} ne "") ? "yes" : "no";
	my $globflags = get_friends_flags($idx, "global");

	Irssi::printformat(MSGLEVEL_CRAP, 'friends_line_head',
		$friends[$idx]->{number},
		$friends[$idx]->{handle},
		(($globflags eq "") ? "[none]" : "$globflags"),
		$passset);

	Irssi::printformat(MSGLEVEL_CRAP, 'friends_line_hosts',
		join(", ", get_friends_hosts($idx)) );

	foreach my $chan (get_friends_channels($idx)) {
		my $flags = get_friends_flags($idx, $chan);
		my $delay = get_friends_delay($idx, $chan);
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_line_chan', 
			$chan,
			(($flags eq "") ? "[none]" : "$flags"),
			(($delay eq "") ? "random" : "$delay"));
	}

	if ($nick ne "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_line_currentnick', $nick);
	}
}

# void add_operation($server, "#channel", "op|voice", timeout, "nick1", "nick2" ..)
# adds a delayed operation
sub add_operation {
	my ($server, $channel, $operation, $timeout, @nicks) = @_;

	# consists of { 
	#  server=>$server,
	#  left=>secondsLeft, 
	#  nicks=>[ nicks ],
	#  channel=>"#channel", 
	#  operation=>"op|voice"
	# }
	push(@operationQueue,
	{
		server=>$server,
		left=>$timeout,
		nicks=>[ @nicks ],
		channel=>$channel,
		operation=>$operation
	});

	debug("add_operation(): Adding operation to queue, left=$timeout, chan=$channel, operation=$operation", 4);

	if (!defined $timerHandle) {
		$timerHandle = Irssi::timeout_add(1000, 'timer_handler', 0);
	}
}

# void timer_handler()
# handles delay timer
sub timer_handler {
	my @ops = ();

	# splice out expired timeouts. if they are expired, move them to
	# local ops-queue. this allows creating new operations to the queue
	# in the operation. (we're not (yet) doing that)

	for (my $c = 0; $c < @operationQueue;) {
		if ($operationQueue[$c]->{left} <= 0) {
			push(@ops, splice(@operationQueue, $c, 1));
		} else {
			++$c;
		}
	}

	for (my $c = 0; $c < @ops; ++$c) {
		my $op = $ops[$c];
		my $channel = $op->{server}->channel_find($op->{channel});

		# check if $channel is still active (you might've parted)
		if ($channel) {
			my @operationNicks = ();
			foreach my $nickStr (@{$op->{nicks}}) {
				my $nick = $channel->nick_find($nickStr);
				# check if there's still such nick (it might've quit/parted)
				if ($nick) {
					# because of a happy coincidence (irssi structures and
					# commands are the same) we can do this:
					if (!$nick->{$op->{operation}}) {
						push(@operationNicks, $nick->{nick});
					}
				}
			}
			# final stage: issue desired command if we're a chanop
			if ($channel->{chanop}) {
				$channel->command($op->{operation}." ".join(" ", @operationNicks));
			}
		}
	}

	# decrement timeouts.
	for (my $c = 0; $c < @operationQueue; ++$c) {
		--$operationQueue[$c]->{left};
		debug("timer_handler(): Decreasing timeout: $operationQueue[$c]->{left} sec left", 5);
	}

	# if operation queue is empty, remove timer.
	if (!@operationQueue) {
		Irssi::timeout_remove($timerHandle);
		$timerHandle = undef;
	}
}

# void load_friends()
# loads friends from file
sub load_friends {
	my $friendfile = Irssi::settings_get_str('friends_file');
	if (-e $friendfile) {
		@friends = ();
		local *F;
		open(F, "<$friendfile");
		local $/ = "\n";
		while (<F>) {
			my ($handle, $hosts, $globflags, $chanstr, $password);
			chop;

			# dealing with comments
			if (/^\#/) {
				# get version of friendlist's author's script
				if (/^\# version = (.+)/) {
					$friends_file_version = $1;
				}
				# get the last /savefriend's timestamp
				if (/^\# written = ([0-9]+)/) {
					$friends_file_written = $1;
				}
				next;
			}

			# split by '%'
			my @entries = split("%", $_);
			foreach my $field (@entries) {
				if ($field =~ /^handle=(.*)$/) {
					$handle = $1;
				} elsif ($field =~ /^hosts=(.*)$/) {
					$hosts = $1;
				} elsif ($field =~ /^globflags=(.*)$/) {
					$globflags = $1;
				} elsif ($field =~ /^chanflags=(.*)$/) {
					$chanstr = $1;
				} elsif ($field =~ /^password=(.*)$/) {
					$password = $1;
				}
			}

			# handle cannot start with a digit
			# skip friend if it does
			if ($handle =~ /^[0-9]/) {
				next;
			}

			# if all fields were processed, and $handle is unique,
			# make a friend and add it to $friends
			debug("load_friends(): calling uniq_handle($handle)", 5);
			if (uniq_handle($handle)) {
				my $newfriend = new_friend($handle, $hosts, $globflags, $chanstr, $password);
				push(@friends, $newfriend);
				debug("load_friends(): Loaded $handle from file", 4);
			} else {
				debug("load_friends(): Duplicate $handle in friends file", 1);
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_duplicate', $handle);
			}
		}

		close(F);

		# check friendlist's version
		my ($verl, $vern) = $friends_file_version =~ /^(.+)\ \(([0-9]+)\)$/;
		debug("load_friends(): checking version gave literal='$verl', numerical='$vern'", 3);
		if ($vern < 20011218) {
			# refuse to load friendlist, if too old.
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_filetooold');
			# clear friendlist
			@friends = ();
			return;
		}

		# if everything's ok -- print a message
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_loaded', scalar(@friends), $friendfile);

	} else {
		# whoops, no such file? bad perms?
		Irssi::print("Cannot load $friendfile");
	}
}

# void cmd_loadfriends($data, $server, $channel)
# handles /loadfriends
sub cmd_loadfriends {
	load_friends();
}

# void save_friends()
# saving friends to file
sub save_friends {
	my $friendfile = Irssi::settings_get_str('friends_file');
	my $current_time = time();
	local *F;
	open(F, ">$friendfile") or die "Couldn't open $friendfile for writing";

	# write script's version and update corresponding variable
	print(F "# version = $friends_version\n");
	$friends_file_version = $friends_version;
	# write current unixtime and update corresponding variable
	print(F "# written = $current_time\n");
	$friends_file_written = $current_time;

	# go through all entries
	for (my $idx = 0; $idx < @friends; ++$idx) {
		# get friend's channels, corresponding flags and delay values
		# then put them as c,f,d fields into @chanstr
		my @chanstr = ();
		foreach my $chan (get_friends_channels($idx)) {
			push(@chanstr, $chan.",".(get_friends_flags($idx, $chan)).",".
				(get_friends_delay($idx, $chan)));
		}

		debug("save_friends(): Writing $friends[$idx]->{handle} to file", 5);

		# write the actual line
		print(F join("%",
			"handle=".$friends[$idx]->{handle},
			"hosts=".$friends[$idx]->{hosts},
			"globflags=".(get_friends_flags($idx, "global")),
			"chanflags=".(join(" ", @chanstr)),
			"password=".$friends[$idx]->{password},
			"\n"));
	}
	# close file, print message
	close(F);
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_saved', scalar(@friends), $friendfile);
}

# void cmd_savefriends($data, $server, $channel)
# handles /savefriends
sub cmd_savefriends {
	eval {
		save_friends();
	};
	if ($?) {
		Irssi::print("Save failed: $?");
	}
}

# void check_friends($server, $channelstr, @nickstocheck)
# checks the given nicklist, channelname and server against the friendlist
sub check_friends {
	my ($server, $channelName, @nicks) = @_;
	my $channel = $server->channel_find($channelName);
	my %opList = ();
	my %voiceList = ();

	debug("check_friends(): $server->{address}, $channelName, nickcount=".scalar(@nicks), 5);

	# get settings
	my $min = Irssi::settings_get_int('friends_delay_min');
	my $max = Irssi::settings_get_int('friends_delay_max');

	# lazy man's sanity checks :-P
	$min = $delay_min if $min < 0;
	$max = $delay_max if $min > $max;

	# for each nick from the given list
	foreach my $nick (@nicks) {
		# check if he's a friend on a given channel
		if ((my $idx = get_idx($channelName, $nick->{nick}, $nick->{host})) > -1) {

			# if so, check his delaytime
			my $delay = get_friends_delay($idx, $channelName);

			# skip this friend if he doesn't have +a
			# (here, we take care about +a'ed people ONLY)
			next if (!friend_has_flag($idx, "a", $channelName) && 
					!friend_has_flag($idx, "a", "global"));

			# notice: password doesn't matter in this loop

			# add $nick to opList{0} if he has global/$channelName +o
			# and isn't opped already
			if ((friend_has_flag($idx, "o", $channelName) || friend_has_flag($idx, "o", "global")) &&
				!($nick->{op})) {
				$opList{0}->{$nick->{nick}} = 1;
			}
			# add $nick to opList{delay} if he has global/$channelName +O
			# and isn't opped already
			if ((friend_has_flag($idx, "O", $channelName) || friend_has_flag($idx, "O", "global")) &&
				!($nick->{op})) {
				# if he doesn't have a delay set for this
				# channel -- calculate one
				if ($delay eq "") {
					$delay = int(rand ($max - $min)) + $min;
				}
				$opList{$delay}->{$nick->{nick}} = 1;
			}
			# add $nick to voiceList{0} if he has global/$channelName +v
			# and isn't voiced already
			if ((friend_has_flag($idx, "v", $channelName) || friend_has_flag($idx, "v", "global")) &&
				!($nick->{voice})) {
				$voiceList{0}->{$nick->{nick}} = 1;
			}
			# add $nick to voiceList{delay} if he has global/$channelName +V
			# and isn't voiced already
			if ((friend_has_flag($idx, "V", $channelName) || friend_has_flag($idx, "V", "global")) &&
				!($nick->{voice})) {
				# if he doesn't have a delay set for this
				# channel -- calculate one
				if ($delay eq "") {
					$delay = int(rand ($max - $min)) + $min;
				}
				$voiceList{$delay}->{$nick->{nick}} = 1;
			}
		}
	}

	# opping
	foreach my $delay (keys %opList) {
		add_operation($server, $channelName, "op", $delay, keys %{$opList{$delay}});
		debug("check_friends(): adding operation: $server->{address}, $channelName, op, $delay", 3);
	}
	# voicing
	foreach my $delay (keys %voiceList) {
		add_operation($server, $channelName, "voice", $delay, keys %{$voiceList{$delay}});
		debug("check_friends(): adding operation: $server->{address}, $channelName, voice, $delay", 3);
	}

	debug("check_friends(): calling timer_handler()", 3);
	timer_handler();
}

# void event_modechange($server, $data, $nick)
# handles modechanges and learning
sub event_modechange {
	my ($server, $data, $nick) = @_;
	my ($channel, $modeStr, $nickStr) = $data =~ /^([^ ]+) ([^ ]+) (.*)$/;
	my @nicks = split(" ", $nickStr);
	my $mode = undef;
	my $nickIndex = 0;
	my $myNick = $server->{nick};
	my $gotOpped = 0;
	my $learnFriends = Irssi::settings_get_bool('friends_learn');
	my $channelInfo = $server->channel_find($channel);
	my ($opperInfo, $opperIdx);

	debug("event_modechange(): Got mode change on $server->{address}: $nick/$channel $modeStr $nickStr", 4);

	# don't learn if learning is disabled, or I am the opper
	if ($learnFriends && ($nick ne $myNick)) {
		if ($channelInfo) {
			$opperInfo = $channelInfo->nick_find($nick);
			$opperIdx = get_idx($channel, $opperInfo->{nick}, $opperInfo->{host});
		} else {
			$opperIdx = -1;
		}
	} else {
		$opperIdx = -1;
	}

	for (my $c = 0; $c < length($modeStr); $c++) {
		if (substr($modeStr, $c, 1) eq "+") {
			$mode = "+";
		} elsif (substr($modeStr, $c, 1) eq "-") {
			$mode = "-";
		} elsif (substr($modeStr, $c, 1) eq "o") {
			if ($mode eq "+" && $nicks[$nickIndex] eq $myNick) {
				$gotOpped = 1;
				debug("event_modechange(): Got opped by $nick", 4);
			}

			if ($opperIdx != -1 && $mode eq "+" && !($gotOpped) &&
				(friend_has_flag($opperIdx, "F", $channel) || 
				 friend_has_flag($opperIdx, "F", 'global'))) {
				my $nickInfo = $channelInfo->nick_find($nicks[$nickIndex]);
				my $friend;
				my $friendIdx = get_idx($channel, $nickInfo->{nick}, $nickInfo->{host});
				debug("event_modechange(): Checked ".$nickInfo->{nick}."!".$nickInfo->{host}." (opped by $nick), got $friendIdx", 4);

				if ($friendIdx == -1) {
					# we got someone not known before
					# choose a handle for him and add him to our friendlist with +L $channel
					my $newHandle = choose_handle($nicks[$nickIndex]);
					$friend = new_friend($newHandle, "*!".$nickInfo->{host}, "", $channel.",L,", "");
					push(@friends, $friend);
					debug("event_modechange(): Added $newHandle to userlist (learnt)", 2);
				} else {
					# we know him already
					$friend = $friends[$friendIdx];
					debug("event_modechange(): dealing with friend == $friend->{handle}", 4);
					# check if his only flag for this channel is 'L"
					if (get_friends_flags($friendIdx, $channel) =~ /^L$/) {
						# if so, add him to the opper's friendlist
						push(@{$friends[$opperIdx]->{friends}}, $friend);
						debug("event_modechange(): added $friend->{handle} to $friends[$opperIdx]->{handle}'s friendlist", 5);
					} else {
						debug("event_modechange(): $friend->{handle} has more than 'L' for $channel, leave him alone", 5);
					}
				}

				if ($friendIdx == -1) {
					debug("event_modechange(): pushing $friend->{handle} to $friends[$opperIdx]->{handle}'s friendlist", 5);
					push(@{$friends[$opperIdx]->{friends}}, $friend);
				}
			}
			++$nickIndex;
		}
	}

	if ($gotOpped && $channelInfo) {
		check_friends($server, $channel, $channelInfo->nicks());
	}
}

# void event_massjoin($channel, $nicklist)
# handles join event
sub event_massjoin {
	my ($channel, $nicksList) = @_;
	my @nicks = @{$nicksList};
	my $server = $channel->{'server'};
	my $channelName = $channel->{name};
	my $begin = time;

	check_friends($server, $channelName, @nicks);

	my $duration = time - $begin;
	if ($duration >= 2) {
		# if checking took more than 2 seconds -- print a message about it
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_checking', $channelName, $duration, $server->{address});
	}
}

# void friends_exec_chflag($idx, $mode, $flag, $chan)
# executes flags altering (changes $friends[$idx])
sub friends_exec_chflag {
	my ($idx, $mode, $flag, $chan) = @_;

	if ($mode eq "+" && is_allowed_flag($flag)) {
		# ADDING flags
		if ($chan eq "" || $chan eq "global") {
			# alter global flags, if $chan is empty or 'global'
			if (!friend_has_flag($idx, $flag, $chan)) {
				# and there's no such flag yet
				addflag($idx, $flag, $chan);
				debug("friends_exec_chflag(): added $flag to $friends[$idx]->{handle}'s global flags", 3);
			}
		} else {
			# ok, so we're playing with channel flags
			if (!friend_has_chanrec($idx, $chan)) {
				# add chan record, if there isn't one
				add_chanrec($idx, $chan);
			}
			if (!friend_has_flag($idx, $flag, $chan)) {
				# add this flag if he doesn't have it yet
				addflag($idx, $flag, $chan);
				debug("friends_exec_chflag(): added $flag to $friends[$idx]->{handle}'s $chan flags", 3);
			}
		}
	} elsif ($mode eq "-") {
		# REMOVING flags
		if ($chan eq "" || $chan eq "global") {
			# global flags: remove $flag from globflags
			delflag($idx, $flag, $chan);
			debug("friends_exec_chflag(): deleted $flag from $friends[$idx]->{handle}'s global flags", 3);
		} else {
			# channel flags: check if $idx has a chanrec for $chan
			if (friend_has_chanrec($idx, $chan)) {
				# if he does, remove $flag from {$chan}->{flags}
				delflag($idx, $flag, $chan);
				debug("friends_exec_chflag(): deleted $flag from $friends[$idx]->{handle}'s $chan flags", 3);
			}
		}
	}
}

# void friends_chflags($idx, $string, $chan)
# parses the $string and calls friends_exec_chflags()
sub friends_chflags {
	my ($idx, $string, $chan) = @_;
	my $wasplus = 0;
	my $wasminus = 0;
	my $char;

	for (my $c = 0; $c < length($string); $c++) {
		$char = substr($string, $c, 1);
		if ($char eq "+") {
			$wasplus = 1;
			$wasminus = 0;
		} elsif ($char eq "-") {
			$wasplus = 0;
			$wasminus = 1;
		} else {
			# commented out: we're gonna check if the flag is valid
			# in friends_exec_chflag(), not here
			# (ability to delete [but NOT add] flags out of range)
			#if (is_allowed_flag($char)) {
				if ($wasplus && !$wasminus) {
					debug("friends_chflags(): Going to friends_exec_chflag($idx, +, $char, $chan)", 5);
					friends_exec_chflag($idx, '+', $char, $chan);
				} elsif (!$wasplus && $wasminus) {
					debug("friends_chflags(): Going to friends_exec_chflag($idx, -, $char, $chan)", 5);
					friends_exec_chflag($idx, '-', $char, $chan);
				}
			#}
		}
	}
}

# void cmd_chflags($data, $server, $channel)
# handles /chflags <handle> <+-flags> [#channel]
sub cmd_chflags {
	my ($handle, $flags, $chans) = split(" ", @_[0], 3);
	my $usage = "/CHFLAGS <handle> <+-flags> [#channel1] [#channel2] ...";

	# remove trailing spaces (just-in-case ;)
	$flags =~ s/[\ ]+$//;
	$chans =~ s/[\ ]+$//;

	# not enough args
	if ($handle eq "" || $flags eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}

	# bad args
	# if the 'flags' part doesn't start with + or -
	if ((substr($flags, 0, 1) ne "+") && (substr($flags, 0, 1) ne "-")) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_badargs', $usage);
		return;
	}

	# get idx, yell and return if it isn't valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	# if #channel wasn't specified -- we'll deal with global flags
	$chans = "global" if ($chans eq "");

	# go through all channels specified
	foreach my $chan (split(/[,\ ]+/, $chans)) {
		debug("cmd_chflags(): going to exec friends_chflags($idx, $flags, '$chan')", 4);

		# 'executing +foo-bar for someone'
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_chflagexec', $flags, $friends[$idx]->{handle}, $chan);
		# make changes
		friends_chflags($idx, $flags, $chan);

		my $flagstr = get_friends_flags($idx, $chan);
		# 'current $chan flags for someone are: +blah/[none]'
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_currentflags', (($flagstr ne "") ? $flagstr : "[none]"), $friends[$idx]->{handle}, $chan);
	}
}

# void cmd_chhandle($data, $server, $channel)
# handles /chhandle <oldhandle> <newhandle>
sub cmd_chhandle {
	my ($oldhandle, $newhandle) = split(" ", @_[0]);
	my $usage = "/CHHANDLE <oldhandle> <newhandle>";

	# remove trailing spaces (just-in-case ;)
	$newhandle =~ s/[\ ]+$//;

	# not enough args
	if ($oldhandle eq "" || $newhandle eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}

	# handle cannot start with a digit
	if ($newhandle =~ /^[0-9]/) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_badhandle', $newhandle, 
			"Handle may not start with a digit");
		return;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($oldhandle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $oldhandle);
		return;
	}

	# check if $newhandle is unique
	# if not, print appropriate message and return
	if (!uniq_handle($newhandle)) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notuniqhandle', $newhandle);
		return;
	}

	# ok, everything seems fine now, let's change the handle.
	$friends[$idx]->{handle} = $newhandle;
	# ... and print a message
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_changed_handle', $oldhandle, $newhandle);
}

# void cmd_chpass($data, $server, $channel)
# handles /chpass <handle> [pass]
# if pass is empty, removes password
# otherwise, crypts it and sets as current one
sub cmd_chpass {
	my ($handle, $pass) = split(" ", @_[0]);
	my $usage = "/CHPASS <handle> [newpassword]";

	# remove trailing space (just-in-case ;)
	$pass =~ s/[\ ]+$//;

	# not enough args
	if ($handle eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	# crypt and set password. then print a message
	$friends[$idx]->{password} = friends_crypt("$pass");
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_chpassexec', $friends[$idx]->{handle});
}

# void cmd_chdelay($data, $server, $channel)
# handles /chdelay <handle> <delay> <#channel>
# if delay is 'remove' or 'none', removes it (thus -- making it random)
# otherwise, sets it as the current one
sub cmd_chdelay {
	my ($handle, $delay, $chan) = split(" ", @_[0]);
	my $usage = "/CHDELAY <handle> <delay> <#channel>";
	my $value;

	# remove trailing space (just-in-case ;)
	$chan =~ s/[\ ]+$//;

	# not enough args
	if ($handle eq "" || $delay eq "" || $chan eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}
	# if $chan doesn't start with one of the [!&#+]
	if (!($chan =~ /^[\!\&\#\+]/)) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_badargs', $usage);
		return;
	}

	# check validness of $delay
	if ($delay =~ /^[0-9]+$/) {
		# numeric value
		$value = $delay;
	} elsif (lc($delay) eq "remove" || lc($delay) eq "none") {
		# 'remove' or 'none'
		$value = "";
	} else {
		# badargs, return
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_badargs', $usage);
		return;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	# check if $idx has got $chan record.
	# add one if needed
	if (!friend_has_chanrec($idx, $chan)) {
		add_chanrec($idx, $chan);
	}

	# finally, set it, and print a message
	chdelay($idx, $value, $chan);
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_changed_delay', $friends[$idx]->{handle},
		$chan, (($value eq "") ? "[random]" : $value));
}

# void cmd_listfriend($searchFor, $server, $chanel)
# handles /listfriends [what]
# 'what' can be either handle, channel name, 1,2,5,15-style or empty.
sub cmd_listfriends {
	if (@friends == 0) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_empty');
	} else {
		my ($searchFor) = @_;
		# remove whitespaces
		$searchFor =~ s/[\ ]+//g;

		if ($searchFor =~ /^[\!\&\#\+]/) {
			# deal with channel
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_friendlist', "channel " . $searchFor);
			for (my $idx = 0; $idx < @friends; ++$idx) {
				listfriend($idx) if (friend_has_chanrec($idx, $searchFor));
			}
		} elsif ($searchFor =~ /^[0-9,]+$/) {
			# deal with 1,2,5,15 style
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_friendlist', $searchFor);
			foreach my $idx (split(/,/, $searchFor)) {
				listfriend($idx) if ($idx < @friends);
			}
		} else {
			if ((my $idx = get_idxbyhand($searchFor)) > -1) {
				# deal with handle
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_friendlist', $searchFor);
				listfriend($idx);
			} else {
				# deal with every entry
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_friendlist', "all");
				for (my $idx = 0; $idx < @friends; ++$idx) {
					listfriend($idx);
				}
			}
		}
	}
}

# void cmd_addfriend($data, $server, $channel)
# handles /addfriend <handle> <hostmask> [flags]
# appends friends_default_flags to 'flags'
sub cmd_addfriend {
	my ($handle, $host, $flags) = split(" ", $_[0]);
	my $usage = "/ADDFRIEND <handle> <hostmask> [flags]";

	# remove whitespaces (just-in-case ;)
	$flags =~ s/[\ ]+//;

	# not enough args
	if ($handle eq "" || $host eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return 0;
	}

	# handle cannot start with a digit
	if ($handle =~ /^[0-9]/) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_badhandle', $handle, 
			"Handle may not start with a digit");
		return;
	}

	# check must be unique
	if (!uniq_handle($handle)) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notuniqhandle', $handle);
		return;
	}

	# check 'flags' parameter
	if ($flags eq "") {
		# append default flags if $flags is empty
		$flags = "+".(Irssi::settings_get_str('friends_default_flags'));
	} elsif (substr($flags, 0, 1) ne "+") {
		# if $flags contains something, check if it starts with '+'
		# if not, prepend one.
		$flags = "+".$flags;
	}

	# add friend.
	debug("cmd_addfriend(): calling new_friend($handle, $host, '', '', '')", 4);
	my $friend = new_friend($handle, $host, "", "", "");
	push(@friends, $friend);
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_added', $handle);

	# our friend should gave $idx == (scalar(@friends) - 1) now, so we'll use it.
	my $idx = scalar(@friends) - 1;

	# add flags.
	debug("cmd_addfriend(): calling friends_chflags($idx, $flags, 'global')", 4);
	friends_chflags($idx, $flags, "global");
	$flags = get_friends_flags($idx, "global");
	if ($flags ne "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_currentflags', $flags, $handle, "global");
	}
}

# void cmd_delfriend($data, $server, $channel)
# handles /delfriend <handle|number>
sub cmd_delfriend {
	my ($who) = split(" ", $_[0]);
	my $usage = "/DELFRIEND <handle|number>";
	my $idx = -1;

	# not enough args
	if ($who eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}

	if ($who =~ /^[0-9]+$/) {
		# query in numeric format, check if it's valid, yell if it's not
		if ($who > -1 && $who < @friends) {
			$idx = $who;
		} else {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $who);
			return;
		}
	} else {
		# query in literal format
		# get idx by handle, yell and return if it's not valid
		$idx = get_idxbyhand($who);
		if ($idx == -1) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $who);
			return;
		}
	}

	# found a valid idx, now delete a friend
	if ((my $deleted = delfriend($idx)) > -1) {
		# renumber the friendlist, then print a message
		renumber_friendlist();
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_removed', $deleted->{handle});
	} else {
		Irssi::print("Whoops. Bug in /delfriend? Report it to shasta\@quasimodo.olsztyn.tpsa.pl please");
	}
}

# void cmd_addhost($data, $server, $channel)
# handles /addhost <handle> <hostmask>
# hostmask may not overlap with any of the current ones
sub cmd_addhost {
	my ($handle, $host) = split(" ", $_[0]);
	my $usage = "/ADDHOST <handle> <hostmask>";

	# remove whitespaces (just-in-case ;)
	$host =~ s/[\ ]+//;

	# not enough args
	if ($handle eq "" || $host eq  "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return 0;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	if (!friend_has_host($idx, $host)) {
		add_host($idx, $host);
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_host_added', $friends[$idx]->{handle}, $host);
	} else {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_host_exists', $friends[$idx]->{handle}, $host);
	}
}

# void cmd_delhost($data, $server, $channel)
# handles /delhost <handle> <hostmask>
# hostmask should be EXACTLY the same as this in $friends[$idx]->{hosts}
sub cmd_delhost {
	my ($handle, $host) = split(" ", $_[0]);
	my $usage = "/DELHOST <handle> <hostmask>";

	# remove whitespaces (just-in-case ;)
	$host =~ s/[\ ]+//;

	# not enough args
	if ($handle eq "" || $host eq  "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return 0;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	my $matched = 0;
	# go through all friend's hosts
	foreach my $hostmask (get_friends_hosts($idx)) {
		if ($hostmask eq $host) {
			# ... removing them if they are the same
			del_host($idx, $host);
			++$matched;
		}
	}

	# print a message
	if ($matched) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_host_removed', $friends[$idx]->{handle}, $host);
	} else {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_host_notexists', $friends[$idx]->{handle}, $host);
	}
	debug("cmd_delhost(): Matched $matched hosts", 4);
}

# void cmd_delchanrec($data, $server, $channel)
# handles /delchanrec <handle> <#channel>
sub cmd_delchanrec {
	my ($handle, $chan) = split(" ", $_[0]);
	my $usage = "/DELCHANREC <handle> <#channel>";

	# remove whitespaces (just-in-case ;)
	$chan =~ s/[\ ]+//;

	# not enough args
	if ($handle eq "" || $chan eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return 0;
	}

	# get idx, yell and return if it's not valid
	my $idx = get_idxbyhand($handle);
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_nosuch', $handle);
		return;
	}

	if (friend_has_chanrec($idx, $chan)) {
		# if friend has such chanrec, delete it. then print a message
		del_chanrec($idx, $chan);
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_chanrec_removed', $friends[$idx]->{handle}, $chan);
	} else {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_chanrec_notexists', $friends[$idx]->{handle}, $chan);
	}
}

# void cmd_findfriends($data, $server, $channel)
# handles /findfriends
# prints all online friends
sub cmd_findfriends {
	# cycle through all channels
	foreach my $channel (Irssi::channels()) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_findfriends', $channel->{name}, $channel->{server});
		# cycle through all nicks on currenct channel
		foreach my $nick ($channel->nicks()) {
			# check if $nick is a friend of us
			if ((my $idx = get_idx($channel->{name}, $nick->{nick}, $nick->{host})) > -1) {
				# if yes, show his entry
				listfriend($idx, $nick->{nick});
			}
		}
	}
}

# void cmd_isfriend($data, $server, $channel)
# handles /isfriend <nick>
sub cmd_isfriend {
	my ($data, $server, $channel) = @_;
	my $usage = "/ISFRIEND <nick>";

	# remove trailing spaces
	$data =~ s/[\ ]+$//;

	# not enough args
	if ($data eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notenoughargs', $usage);
		return;
	}

	# redirect userhost reply to event_isfriend_userhost()
	# caution: This works with Irssi 0.7.98.CVS (20011117) and newer
	$server->redirect_event("userhost", 1, $data, 0, undef, {
				"event 302" => "redir userhost_friends"});
	debug("cmd_isfriend(): Redirect set up: event 302 -> userhost_friends(), $data", 3);
	# send our query
	$server->send_raw("USERHOST :$data");
	debug("cmd_isfriend(): USERHOST :$data sent to $server->{address}", 3);
}

# void event_isfriend_userhost
# handles redirected USERHOST replies
# (part of /isfriend)
sub event_isfriend_userhost {
	my ($mynick, $reply) = split(/ +/, $_[1]);
	my ($nick, $user, $userhost) = $reply =~ /^:?([^\*=]*)\*?=.(.*)@(.*)/;
	my $friend_matched = 0;

	debug("event_isfriend_userhost(): Got reply: $reply", 3);

	# cycle through all friendlist's entries
	for (my $idx = 0; $idx < @friends; ++$idx) {
		# check if received nick!user@host matches any of our friends
		if (friend_has_host($idx, ($nick."!".$user."@".$userhost))) {
			# if so, give informations about him
			listfriend($idx, $nick);
			$friend_matched = 1;
		}
		# you can uncomment this to stop matching
		# after first successful one:
		#last if ($friend_matched);
	}

	# print a message
	if ($friend_matched) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_endof', "/isfriend", $nick);
	} else {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_notafriend', $nick);
	}
}

# void flushlearnt()
# cycles through all users and removes every chanrec with flag L
# clears the opping tree too
sub flushlearnt {
	# cycle through the whole friendlist
	for (my $idx = 0; $idx < @friends; ++$idx) {
		# clear the opping tree
		$friends[$idx]->{friends} = [];
		# now go through all friend's channel entries
		foreach my $chan (get_friends_channels($idx)) {
			# if 'L' is the only flag for this chan
			if (get_friends_flags($idx, $chan) =~ /^L$/) {
				# remove channel record and print a message
				del_chanrec($idx, $chan);
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_chanrec_removed', $friends[$idx]->{handle}, $chan);
			}
		}
	}
}

# void opping_tree()
# prints the Opping Tree
sub opping_tree {
	Irssi::printformat(MSGLEVEL_CRAP, 'friends_optree_header');
	# cycle through the whole friendlist
	for (my $idx = 0; $idx < @friends; ++$idx) {
		# get friend's friends
		my @friendFriends = @{$friends[$idx]->{friends}};
		if (@friendFriends) {
			# print info about our friend
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_optree_line1', $friends[$idx]->{handle});
			my %masks;
			# get all masks opped by him
			foreach my $friend (@friendFriends) {
				$masks{%{$friend}->{hosts}}++;
			}
			# print them, along with the opcount
			foreach my $friend (sort keys %masks) {
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_optree_line2', $masks{$friend}, $friend);
			}
		}
	}
}

# void cmd_oppingtree($data, $server, $channel)
# handles /oppingtree
sub cmd_oppingtree {
	opping_tree();
}

# void cmd_flushlearnt(data, $server, $channel)
# handles /flushlearnt
sub cmd_flushlearnt {
	flushlearnt();
}

# void event_ctcpmsg($server, $args, $sender, $senderhsot, $target)
# handles ctcp requests
sub event_ctcpmsg {
	my ($server, $args, $sender, $userhost, $target) = @_;
	my $ctcp_sigstop = 0;
	my $idx = -1;

	debug("event_ctcpmsg(): Got CTCP from $sender!$userhost for $target/$server->{address} - $args", 6);

	# return, if ctcp is not for us
	my $myNick = $server->{nick};
	return if (lc($target) ne lc($myNick));

	# return, if we don't accept ctcp requests
	return unless (Irssi::settings_get_bool('friends_use_ctcp'));

	my @cmdargs = split(" ", $args);

	# prepare arguments:

	# get 1st arg, uppercase it
	my $command = uc($cmdargs[0]);

	# get 2nd arg, remove trailing spaces in case of tab-completing or something
	my $channelName = $cmdargs[1];
	$channelName =~ s/[\ ]+$//;

	# get 3rd arg, remove trailing spaces in case of tab-completing or something
	my $password = $cmdargs[2];
	$password =~ s/[\ ]+$//;

	# check if $command is one of friends_ctcp_commands. return if it isn't
	debug("event_ctcpmsg(): calling is_ctcp_command($command)", 5);
	if (!is_ctcp_command($command)) {
		return;
	} else {
		$ctcp_sigstop = 1;
	}

	if ($command eq "PASS") {
		# PASS is weird a bit, it doesn't require friend to have a chanrec or flags at all
		debug("event_ctcpmsg(): calling get_idxbyhost($sender, $userhost)", 5);
		$idx = get_idxbyhost($sender, $userhost);
	} else {
		# Ok, it's NOT a password request.
		# get idx
		debug("event_ctcpmsg(): calling get_idx($channelName, $sender, $userhost)", 5);
		$idx = get_idx($channelName, $sender, $userhost);
	}

	# if get_idx* failed, return.
	if ($idx == -1) {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $sender."!".$userhost, "Not a friend");
		Irssi::signal_stop();
		return;
	}

	# we'll use handle instead of $sender!$userhost in messages
	my $handle = $friends[$idx]->{handle};

	# check if $channelName was supplied.
	# (first argument, should be always given)
	if ($channelName eq "") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough arguments");
		Irssi::signal_stop();
		return;
	}

	# get channel object. if not found -- yell, stop the signal, and return
	debug("event_ctcpmsg(): calling server->channel_find($channelName)", 5);
	my $channel = $server->channel_find($channelName);
	if (!$channel && $command ne "PASS") {
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not on channel $channelName");
		Irssi::signal_stop();
		return;
	}

	debug("event_ctcpmsg(): CTCP from $handle: $command $channelName $password", 5);

	# /ctcp nick OP #channel password
	if ($command eq "OP") {
		if (!(friend_has_flag($idx, "[oO]", $channelName) || friend_has_flag($idx, "[oO]", "global"))) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough flags");
			Irssi::signal_stop();
			return;
		}
		if (!friends_passwdok($idx, $password)) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
			Irssi::signal_stop();
			return;
		}
		# process allowed opping
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcprequest', $handle, $command, $channelName);
		# op that begger!
		debug("event_ctcpmsg(): calling channel->command('op $sender')", 5);
		$channel->command("op $sender");

	# /ctcp nick VOICE #channel password
	} elsif ($command eq "VOICE") {
		if (!(friend_has_flag($idx, "[vV]", $channelName) || friend_has_flag($idx, "[vV]", "global"))) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough flags");
			Irssi::signal_stop();
			return;
		}
		if (!friends_passwdok($idx, $password)) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
			Irssi::signal_stop();
			return;
		}

		# process allowed voicing
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcprequest', $handle, $command, $channelName);
		# voice that begger!
		debug("event_ctcpmsg(): calling channel->command('voice $sender')", 5);
		$channel->command("voice $sender");

	# /ctcp nick INVITE #channel password
	} elsif ($command eq "INVITE") {
		if (!(friend_has_flag($idx, "i", $channelName) || friend_has_flag($idx, "i", "global"))) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough flags");
			Irssi::signal_stop();
			return;
		}
		if (!friends_passwdok($idx, $password)) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
			Irssi::signal_stop();
			return;
		}

		# process allowed invite
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcprequest', $handle, $command, $channelName);
		# invite that begger!
		debug("event_ctcpmsg(): calling channel->command('invite $sender')", 5);
		$channel->command("invite $sender");

	# /ctcp nick KEY #channel password
	} elsif ($command eq "KEY") {
		if (!(friend_has_flag($idx, "k", $channelName) || friend_has_flag($idx, "k", "global"))) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough flags");
			Irssi::signal_stop();
			return;
		}
		if (!friends_passwdok($idx, $password)) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
			Irssi::signal_stop();
			return;
		}

		# process allowed key giving
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcprequest', $handle, $command, $channelName);
		if ($channel->{key} ne '') {
			# give him/her a key!
			debug("event_ctcpmsg(): calling server->command('/^notice $sender key for $channelName is: $channel->{key}')", 5);
			$server->command("/^NOTICE $sender Key for $channelName is: $channel->{key}");
		}

	# /ctcp nick LIMIT #channel password
	} elsif ($command eq "LIMIT") {
		if (!(friend_has_flag($idx, "l", $channelName) || friend_has_flag($idx, "l", "global"))) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Not enough flags");
			Irssi::signal_stop();
			return;
		}
		if (!friends_passwdok($idx, $password)) {
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
			Irssi::signal_stop();
			return;
		}

		# process allowed limit raising
		Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcprequest', $handle, $command, $channelName);
		my @nicks = $channel->nicks();
		if ($channel->{limit} <= scalar(@nicks) && $channel->{limit}) {
			# raise the limit!
			debug("event_ctcpmsg(): calling server->command('MODE $channelName +l ".(scalar(@nicks) + 1)."')", 5);
			$server->command("MODE $channelName +l " . (scalar(@nicks) + 1));
		}

	# /ctcp nick PASS pass [newpass]
	} elsif ($command eq "PASS") {
		# if someone has password already set - we can only *change* it
		if ($friends[$idx]->{password} ne "") {
			# if cmdargs[1] ($channelName, that is) is a valid password (current)
			if (!friends_passwdok($idx, $channelName)) {
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcpfail', $command, $handle, "Bad password");
				Irssi::signal_stop();
				return;
			}
			# and cmdargs[2] ($password, that is) contains something ...
			if ($password ne "") {
				# ... process allowed password change.
				# in this case, old password is in $channelName
				# and new password is in $password
				debug("event_ctcpmsg(): changing password for $handle", 5);
				$friends[$idx]->{password} = friends_crypt("$password");
				Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcppass', $handle, $sender."!".$userhost);
				# send a quiet notice to sender
				$server->command("/^NOTICE $sender Password changed to: $password");
			} else {
				# in this case, notify sender about his current password quietly
				debug("event_ctcpmsg(): Password already set for $handle while trying to change", 4);
				$server->command("/^NOTICE $sender You already have a password set");
			}
		# if $idx doesn't have a password, we will *set* it
		} else {
			# in this case, new password is in $channelName
			# and $password is unused
			debug("event_ctcpmsg(): setting password for $handle", 5);
			$friends[$idx]->{password} = friends_crypt("$channelName");
			Irssi::printformat(MSGLEVEL_CRAP, 'friends_ctcppass', $handle, $sender.'!'.$userhost);
			# send a quiet notice to sender
			$server->command("/^NOTICE $sender Password set to: $channelName");
		}
	}

	# stop the signal if we processed the request
	if ($ctcp_sigstop) {
		debug("event_ctcpmsg(): stopping signal", 4);
		Irssi::signal_stop();
	}
}

# void cmd_friendsversion()
# handles /friendsversion
# prints script's and friendlist's version
sub cmd_friendsversion() {
	printversion("script");
	printversion("filever");
	printversion("filewritten");
}

Irssi::settings_add_int("misc", "friends_delay_min", $delay_min);
Irssi::settings_add_int("misc", "friends_delay_max", $delay_max);
Irssi::settings_add_int("misc", "friends_debug_level", $friends_debug_level);
Irssi::settings_add_bool("misc", "friends_learn", $friends_learn);
Irssi::settings_add_bool("misc", "friends_use_ctcp", $friends_use_ctcp);
Irssi::settings_add_str("misc", "friends_ctcp_commands", $friends_ctcp_commands);
Irssi::settings_add_str("misc", "friends_allowed_flags", $friends_allowed_flags);
Irssi::settings_add_str("misc", "friends_default_flags", $friends_default_flags);
Irssi::settings_add_str("misc", "friends_file", $friends_file);
Irssi::command_bind("addfriend", "cmd_addfriend");
Irssi::command_bind("delfriend", "cmd_delfriend");
Irssi::command_bind("addhost", "cmd_addhost");
Irssi::command_bind("delhost", "cmd_delhost");
Irssi::command_bind("delchanrec", "cmd_delchanrec");
Irssi::command_bind("chhandle", "cmd_chhandle");
Irssi::command_bind("chdelay", "cmd_chdelay");
Irssi::command_bind("loadfriends", "cmd_loadfriends");
Irssi::command_bind("savefriends", "cmd_savefriends");
Irssi::command_bind("listfriends", "cmd_listfriends");
Irssi::command_bind("findfriends", "cmd_findfriends");
Irssi::command_bind("isfriend", "cmd_isfriend");
Irssi::command_bind("chflags", "cmd_chflags");
Irssi::command_bind("chpass", "cmd_chpass");
Irssi::command_bind("oppingtree", "cmd_oppingtree");
Irssi::command_bind("flushlearnt", "cmd_flushlearnt");
Irssi::command_bind("friendsversion", "cmd_friendsversion");
Irssi::signal_add("default ctcp msg", "event_ctcpmsg");
Irssi::signal_add("redir userhost_friends", "event_isfriend_userhost");
Irssi::signal_add_last("massjoin", "event_massjoin");
Irssi::signal_add_last("event mode", "event_modechange");

load_friends();
printversion("script");
